/*
 * livepatch_bsc1097108
 *
 * Fix for CVE-2018-10853, bsc#1097108
 *
 *  Upstream commits:
 *  79367a657439 ("KVM: x86: introduce linear_{read,write}_system")
 *  ce14e868a54e ("KVM: x86: pass kvm_vcpu to kvm_read_guest_virt and
 *                 kvm_write_guest_virt_system")
 *  3c9fa24ca7c9 ("kvm: x86: use correct privilege level for
 *                 sgdt/sidt/fxsave/fxrstor access")
 *
 *  SLE12 commit:
 *  none yet
 *
 *  SLE12-SP1 commit
 *  none yet
 *
 *  SLE12-SP2 commit:
 *  none yet
 *
 *  SLE12-SP3 commit:
 *  06238a187fb7aafb7088fc617cb0a0efd6259d80 (stable 4.4.138)
 *
 *  SLE15 commit:
 *  none yet
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */


#if IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/kvm_host.h>
#include <asm/kvm_emulate.h>
#include <asm/fpu/types.h>
#include <asm/desc_defs.h>
#include "livepatch_bsc1097108.h"

#if !IS_MODULE(CONFIG_KVM)
#error "Live patch supports only CONFIG_KVM=m"
#endif

#define KLP_PATCHED_MODULE "kvm"


static struct kvm_x86_ops **klp_kvm_x86_ops;

static int (*klp_kvm_read_guest_virt_system)(struct x86_emulate_ctxt *ctxt,
					     gva_t addr, void *val,
					     unsigned int bytes,
					     struct x86_exception *exception);
static int (*klp_kvm_write_guest_virt_system)(struct x86_emulate_ctxt *ctxt,
					      gva_t addr, void *val,
					      unsigned int bytes,
					      struct x86_exception *exception);
static int (*klp_kvm_read_guest_virt_helper)(gva_t addr, void *val,
					     unsigned int bytes,
					     struct kvm_vcpu *vcpu, u32 access,
					     struct x86_exception *exception);
static int (*klp_kvm_vcpu_write_guest)(struct kvm_vcpu *vcpu, gpa_t gpa,
				       const void *data, unsigned long len);
static int (*klp_linearize)(struct x86_emulate_ctxt *ctxt,
			    struct segmented_address addr,
			    unsigned size, bool write,
			    ulong *linear);

static struct {
	char *name;
	void **addr;
} klp_funcs[] = {
	{ "kvm:kvm_x86_ops", (void *)&klp_kvm_x86_ops },
	{ "kvm:kvm_read_guest_virt_system",
				(void *)&klp_kvm_read_guest_virt_system },
	{ "kvm:kvm_write_guest_virt_system",
				(void *)&klp_kvm_write_guest_virt_system },
	{ "kvm:kvm_read_guest_virt_helper",
				(void *)&klp_kvm_read_guest_virt_helper },
	{ "kvm:kvm_vcpu_write_guest", (void *)&klp_kvm_vcpu_write_guest },
	{ "kvm:linearize", (void *)&klp_linearize },
};



/* from arch/x86/kvm/x86.c */
#define klp_emul_to_vcpu(ctxt)					\
	container_of(ctxt, struct kvm_vcpu, arch.emulate_ctxt)



/* New */
static int klp_emulator_read_std(struct x86_emulate_ctxt *ctxt,
				 gva_t addr, void *val, unsigned int bytes,
				 struct x86_exception *exception, bool system)
{
	struct kvm_vcpu *vcpu = klp_emul_to_vcpu(ctxt);
	u32 access = 0;

	if (!system && (*klp_kvm_x86_ops)->get_cpl(vcpu) == 3)
		access |= PFERR_USER_MASK;

	return klp_kvm_read_guest_virt_helper(addr, val, bytes, vcpu, access,
					  exception);
}

/* New */
static int klp_kvm_write_guest_virt_helper(gva_t addr, void *val,
					   unsigned int bytes,
					   struct kvm_vcpu *vcpu, u32 access,
					   struct x86_exception *exception)
{
	void *data = val;
	int r = X86EMUL_CONTINUE;

	while (bytes) {
		gpa_t gpa =  vcpu->arch.walk_mmu->gva_to_gpa(vcpu, addr,
							     access,
							     exception);
		unsigned offset = addr & (PAGE_SIZE-1);
		unsigned towrite = min(bytes, (unsigned)PAGE_SIZE - offset);
		int ret;

		if (gpa == UNMAPPED_GVA)
			return X86EMUL_PROPAGATE_FAULT;
		ret = klp_kvm_vcpu_write_guest(vcpu, gpa, data, towrite);
		if (ret < 0) {
			r = X86EMUL_IO_NEEDED;
			goto out;
		}

		bytes -= towrite;
		data += towrite;
		addr += towrite;
	}
out:
	return r;
}

/* New */
static int klp_emulator_write_std(struct x86_emulate_ctxt *ctxt, gva_t addr,
				  void *val, unsigned int bytes,
				  struct x86_exception *exception, bool system)
{
	struct kvm_vcpu *vcpu = klp_emul_to_vcpu(ctxt);
	u32 access = PFERR_WRITE_MASK;

	if (!system && (*klp_kvm_x86_ops)->get_cpl(vcpu) == 3)
		access |= PFERR_USER_MASK;

	return klp_kvm_write_guest_virt_helper(addr, val, bytes, vcpu,
					       access, exception);
}

/* patched */
int klp_segmented_read_std(struct x86_emulate_ctxt *ctxt,
			   struct segmented_address addr,
			   void *data,
			   unsigned size)
{
	int rc;
	ulong linear;

	rc = klp_linearize(ctxt, addr, size, false, &linear);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	/*
	 * Fix CVE-2018-10853
	 *  +5 lines
	 *
	 * Note that ->read_std should always equal
	 * kvm_read_guest_virt_system and that the check below is
	 * only there as a safety measure.
	 */
	if (likely(ctxt->ops->read_std == klp_kvm_read_guest_virt_system)) {
		return klp_emulator_read_std(ctxt, linear, data, size,
					     &ctxt->exception, false);
	}

	return ctxt->ops->read_std(ctxt, linear, data, size, &ctxt->exception);
}

/* patched */
int klp_segmented_write_std(struct x86_emulate_ctxt *ctxt,
			    struct segmented_address addr,
			    void *data,
			    unsigned int size)
{
	int rc;
	ulong linear;

	rc = klp_linearize(ctxt, addr, size, true, &linear);
	if (rc != X86EMUL_CONTINUE)
		return rc;
	/*
	 * Fix CVE-2018-10853
	 *  +5 lines
	 *
	 * Note that ->write_std should always equal
	 * kvm_write_guest_virt_system and that the check below is
	 * only there as a safety measure.
	 */
	if (likely(ctxt->ops->write_std == klp_kvm_write_guest_virt_system)) {
		return klp_emulator_write_std(ctxt, linear, data, size,
					      &ctxt->exception, false);
	}

	return ctxt->ops->write_std(ctxt, linear, data, size, &ctxt->exception);
}



static int livepatch_bsc1097108_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(klp_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(klp_funcs[i].name);
		if (!addr) {
			pr_err("livepatch: symbol %s not resolved\n",
				klp_funcs[i].name);
			return -ENOENT;
		}

		*(klp_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int livepatch_bsc1097108_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KLP_PATCHED_MODULE))
		return 0;

	ret = livepatch_bsc1097108_kallsyms();
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1097108_module_nb = {
	.notifier_call = livepatch_bsc1097108_module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1097108_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KLP_PATCHED_MODULE)) {
		ret = livepatch_bsc1097108_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&livepatch_bsc1097108_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1097108_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1097108_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM) */
