/*
 * livepatch_bsc1099306_kvm_intel
 *
 * Fix for CVE-2018-3646 (kvm_intel.ko part), bsc#1099306
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/kvm_host.h>
#include <asm/vmx.h>
#include <linux/highmem.h>
#include <asm/segment.h>
#include <asm/nospec-branch.h>
#include <linux/frame.h>
#include "livepatch_bsc1099306_kvm_intel.h"
#include "bsc1099306.h"

#if !IS_MODULE(CONFIG_KVM_INTEL)
#error "Live patch supports only CONFIG_KVM_INTEL=m"
#endif

#define KLP_PATCHED_MODULE "kvm_intel"


struct vcpu_vmx;
struct vmcs12;

static struct kvm_x86_ops **klp_kvm_x86_ops;
static bool *klp_kvm_rebooting;
static asmlinkage void (*klp_kvm_spurious_fault)(void);
static int (*klp_kvm_set_shared_msr)(unsigned slot, u64 value, u64 mask);
static void (*klp_kvm_queue_exception)(struct kvm_vcpu *vcpu, unsigned nr);
static bool (*klp_kvm_valid_efer)(struct kvm_vcpu *vcpu, u64 efer);
static int (*klp_kvm_vcpu_halt)(struct kvm_vcpu *vcpu);

static bool *klp_enable_ept;
static bool *klp_enable_shadow_vmcs;
static void (*klp_vmwrite_error)(unsigned long field, unsigned long value);
static u32 (*klp_vmx_get_interrupt_shadow)(struct kvm_vcpu *vcpu);
static void (*klp_nested_vmx_failInvalid)(struct kvm_vcpu *vcpu);
static void (*klp_copy_shadow_to_vmcs12)(struct vcpu_vmx *vmx);
static void (*klp_nested_vmx_failValid)(struct kvm_vcpu *vcpu,
					u32 vm_instruction_error);
static int (*klp_nested_vmx_check_msr_switch)(struct kvm_vcpu *vcpu,
					      unsigned long count_field,
					      unsigned long addr_field);
static void (*klp_skip_emulated_instruction)(struct kvm_vcpu *vcpu);
static void (*klp_load_vmcs12_host_state)(struct kvm_vcpu *vcpu,
					  struct vmcs12 *vmcs12);
static void (*klp_nested_vmx_succeed)(struct kvm_vcpu *vcpu);
static int (*klp_enter_vmx_non_root_mode)(struct kvm_vcpu *vcpu,
					  bool from_vmentry);
static int (*klp_kvm_skip_emulated_instruction)(struct kvm_vcpu *vcpu);
static int (*klp_vmx_get_cpl)(struct kvm_vcpu *vcpu);

static struct {
	char *name;
	void **addr;
} klp_funcs[] = {
	{ "kvm:kvm_x86_ops", (void *)&klp_kvm_x86_ops },
	{ "kvm:kvm_rebooting", (void *)&klp_kvm_rebooting },
	{ "kvm:kvm_spurious_fault", (void *)&klp_kvm_spurious_fault },
	{ "kvm:kvm_set_shared_msr", (void *)&klp_kvm_set_shared_msr },
	{ "kvm:kvm_queue_exception", (void *)&klp_kvm_queue_exception },
	{ "kvm:kvm_valid_efer", (void *)&klp_kvm_valid_efer },
	{ "kvm:kvm_vcpu_halt", (void *)&klp_kvm_vcpu_halt },
	{ "kvm:kvm_skip_emulated_instruction",
			(void *)&klp_kvm_skip_emulated_instruction },

	{ "kvm_intel:enable_ept", (void *)&klp_enable_ept },
	{ "kvm_intel:enable_shadow_vmcs", (void *)&klp_enable_shadow_vmcs },
	{ "kvm_intel:vmwrite_error", (void *)&klp_vmwrite_error },
	{ "kvm_intel:vmx_get_interrupt_shadow",
			(void *)&klp_vmx_get_interrupt_shadow },
	{ "kvm_intel:nested_vmx_failInvalid",
			(void *)&klp_nested_vmx_failInvalid },
	{ "kvm_intel:copy_shadow_to_vmcs12",
			(void *)&klp_copy_shadow_to_vmcs12 },
	{ "kvm_intel:nested_vmx_failValid",
			(void *)&klp_nested_vmx_failValid },
	{ "kvm_intel:nested_vmx_check_msr_switch",
			(void *)&klp_nested_vmx_check_msr_switch },
	{ "kvm_intel:skip_emulated_instruction",
			(void *)&klp_skip_emulated_instruction },
	{ "kvm_intel:load_vmcs12_host_state",
			(void *)&klp_load_vmcs12_host_state },
	{ "kvm_intel:nested_vmx_succeed", (void *)&klp_nested_vmx_succeed },
	{ "kvm_intel:enter_vmx_non_root_mode",
			(void *)&klp_enter_vmx_non_root_mode },
	{ "kvm_intel:vmx_get_cpl", (void *)&klp_vmx_get_cpl },
};


/* from arch/x86/include/asm/kvm_host.h */
/*
 * Avoid module dependency on kvm.ko.
 *
 * This is the original macro but with the load from kvm_rebooting
 * replaced by an indirect load from *klp_kvm_rebooting and the call
 * to kvm_spurious_fault() replaced by an indirect call to
 * *klp_kvm_spurious_fault.
 */
#define klp____kvm_handle_fault_on_reboot(insn, cleanup_insn)	\
	"666: " insn "\n\t"				\
	"668: \n\t"					\
	".pushsection .fixup, \"ax\" \n"		\
	"667: \n\t"					\
	cleanup_insn "\n\t"				\
	"pushq %%rax \n\t"				\
	"movq klp_kvm_rebooting, %%rax\n\t"		\
	"cmpb $0, (%%rax) \n\t"			\
	"popq %%rax \n\t"				\
	"jne 668b \n\t"				\
	__ASM_SIZE(push) " $666b \n\t"			\
	/* this depends on [thunk_target] "a" constraints */	\
	"movq klp_kvm_spurious_fault, %%rax \n\t"	\
	CALL_NOSPEC "\n\t"				\
	/* kvm_spurious_fault() is noreturn. */	\
	/* Silence objtool by balancing the stack */	\
	/* and jmp'ing somewhere */			\
	"popq %%rax\n\t"				\
	"jmp 668b \n\t"				\
	".popsection \n\t"				\
	_ASM_EXTABLE(666b, 667b)

#define klp__kvm_handle_fault_on_reboot(insn)		\
	klp____kvm_handle_fault_on_reboot(insn, "")

/* from arch/x86/kvm/kvm_cache_regs.h */
#define KLP_KVM_POSSIBLE_CR4_GUEST_BITS				  \
	(X86_CR4_PVI | X86_CR4_DE | X86_CR4_PCE | X86_CR4_OSFXSR  \
	 | X86_CR4_OSXMMEXCPT | X86_CR4_LA57 | X86_CR4_PGE)

/* inlined */
static inline ulong klp_kvm_read_cr4_bits(struct kvm_vcpu *vcpu, ulong mask)
{
	ulong tmask = mask & KLP_KVM_POSSIBLE_CR4_GUEST_BITS;
	if (tmask & vcpu->arch.cr4_guest_owned_bits)
		(*klp_kvm_x86_ops)->decache_cr4_guest_bits(vcpu);
	return vcpu->arch.cr4 & mask;
}


/* from arch/x86/kvm/x86.h */
/* inlined */
static inline int klp_is_long_mode(struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return vcpu->arch.efer & EFER_LMA;
#else
	return 0;
#endif
}

/* inlined */
static inline u8 klp_vcpu_virt_addr_bits(struct kvm_vcpu *vcpu)
{
	return klp_kvm_read_cr4_bits(vcpu, X86_CR4_LA57) ? 57 : 48;
}

/* inlined */
static inline u64 klp_get_canonical(u64 la, u8 vaddr_bits)
{
	return ((int64_t)la << (64 - vaddr_bits)) >> (64 - vaddr_bits);
}

/* inlined */
static inline bool klp_is_noncanonical_address(u64 la, struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return klp_get_canonical(la, klp_vcpu_virt_addr_bits(vcpu)) != la;
#else
	return false;
#endif
}


/* from arch/x86/kvm/cpuid.h */
/* inlined */
static inline int klp_cpuid_maxphyaddr(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.maxphyaddr;
}


/* from arch/x86/kvm/vmx.c */
#define klp__ex(x) klp__kvm_handle_fault_on_reboot(x)
#define klp__ex_clear(x, reg) \
	klp____kvm_handle_fault_on_reboot(x, "xor " reg " , " reg)

#define KLP_NR_AUTOLOAD_MSRS 8

struct shared_msr_entry {
	unsigned index;
	u64 data;
	u64 mask;
};

struct loaded_vmcs {
	struct vmcs *vmcs;
	struct vmcs *shadow_vmcs;
	int cpu;
	bool launched;
	bool nmi_known_unmasked;
	unsigned long vmcs_host_cr3;	/* May not match real cr3 */
	unsigned long vmcs_host_cr4;	/* May not match real cr4 */
	/* Support for vnmi-less CPUs */
	int soft_vnmi_blocked;
	ktime_t entry_time;
	s64 vnmi_blocked_time;
	unsigned long *msr_bitmap;
	struct list_head loaded_vmcss_on_cpu_link;
};

struct nested_vmx {
	/* Has the level1 guest done vmxon? */
	bool vmxon;
	gpa_t vmxon_ptr;
	bool pml_full;

	/* The guest-physical address of the current VMCS L1 keeps for L2 */
	gpa_t current_vmptr;
	/*
	 * Cache of the guest's VMCS, existing outside of guest memory.
	 * Loaded from guest memory during VMPTRLD. Flushed to guest
	 * memory during VMCLEAR and VMPTRLD.
	 */
	struct vmcs12 *cached_vmcs12;
	/*
	 * Indicates if the shadow vmcs must be updated with the
	 * data hold by vmcs12
	 */
	bool sync_shadow_vmcs;

	bool change_vmcs01_virtual_x2apic_mode;
	/* L2 must run next, and mustn't decide to exit to L1. */
	bool nested_run_pending;

	struct loaded_vmcs vmcs02;

	/*
	 * Guest pages referred to in the vmcs02 with host-physical
	 * pointers, so we must keep them pinned while L2 runs.
	 */
	struct page *apic_access_page;
	struct page *virtual_apic_page;
	struct page *pi_desc_page;
	struct pi_desc *pi_desc;
	bool pi_pending;
	u16 posted_intr_nv;

	struct hrtimer preemption_timer;
	bool preemption_timer_expired;

	/* to migrate it to L2 if VM_ENTRY_LOAD_DEBUG_CONTROLS is off */
	u64 vmcs01_debugctl;

	u16 vpid02;
	u16 last_vpid;

	/*
	 * We only store the "true" versions of the VMX capability MSRs. We
	 * generate the "non-true" versions by setting the must-be-1 bits
	 * according to the SDM.
	 */
	u32 nested_vmx_procbased_ctls_low;
	u32 nested_vmx_procbased_ctls_high;
	u32 nested_vmx_secondary_ctls_low;
	u32 nested_vmx_secondary_ctls_high;
	u32 nested_vmx_pinbased_ctls_low;
	u32 nested_vmx_pinbased_ctls_high;
	u32 nested_vmx_exit_ctls_low;
	u32 nested_vmx_exit_ctls_high;
	u32 nested_vmx_entry_ctls_low;
	u32 nested_vmx_entry_ctls_high;
	u32 nested_vmx_misc_low;
	u32 nested_vmx_misc_high;
	u32 nested_vmx_ept_caps;
	u32 nested_vmx_vpid_caps;
	u64 nested_vmx_basic;
	u64 nested_vmx_cr0_fixed0;
	u64 nested_vmx_cr0_fixed1;
	u64 nested_vmx_cr4_fixed0;
	u64 nested_vmx_cr4_fixed1;
	u64 nested_vmx_vmcs_enum;
	u64 nested_vmx_vmfunc_controls;

	/* SMM related state */
	struct {
		/* in VMX operation on SMM entry? */
		bool vmxon;
		/* in guest mode on SMM entry? */
		bool guest_mode;
	} smm;
};

typedef u64 natural_width;
struct __packed vmcs12 {
	/* According to the Intel spec, a VMCS region must start with the
	 * following two fields. Then follow implementation-specific data.
	 */
	u32 revision_id;
	u32 abort;

	u32 launch_state; /* set to 0 by VMCLEAR, to 1 by VMLAUNCH */
	u32 padding[7]; /* room for future expansion */

	u64 io_bitmap_a;
	u64 io_bitmap_b;
	u64 msr_bitmap;
	u64 vm_exit_msr_store_addr;
	u64 vm_exit_msr_load_addr;
	u64 vm_entry_msr_load_addr;
	u64 tsc_offset;
	u64 virtual_apic_page_addr;
	u64 apic_access_addr;
	u64 posted_intr_desc_addr;
	u64 vm_function_control;
	u64 ept_pointer;
	u64 eoi_exit_bitmap0;
	u64 eoi_exit_bitmap1;
	u64 eoi_exit_bitmap2;
	u64 eoi_exit_bitmap3;
	u64 eptp_list_address;
	u64 xss_exit_bitmap;
	u64 guest_physical_address;
	u64 vmcs_link_pointer;
	u64 pml_address;
	u64 guest_ia32_debugctl;
	u64 guest_ia32_pat;
	u64 guest_ia32_efer;
	u64 guest_ia32_perf_global_ctrl;
	u64 guest_pdptr0;
	u64 guest_pdptr1;
	u64 guest_pdptr2;
	u64 guest_pdptr3;
	u64 guest_bndcfgs;
	u64 host_ia32_pat;
	u64 host_ia32_efer;
	u64 host_ia32_perf_global_ctrl;
	u64 padding64[8]; /* room for future expansion */
	/*
	 * To allow migration of L1 (complete with its L2 guests) between
	 * machines of different natural widths (32 or 64 bit), we cannot have
	 * unsigned long fields with no explict size. We use u64 (aliased
	 * natural_width) instead. Luckily, x86 is little-endian.
	 */
	natural_width cr0_guest_host_mask;
	natural_width cr4_guest_host_mask;
	natural_width cr0_read_shadow;
	natural_width cr4_read_shadow;
	natural_width cr3_target_value0;
	natural_width cr3_target_value1;
	natural_width cr3_target_value2;
	natural_width cr3_target_value3;
	natural_width exit_qualification;
	natural_width guest_linear_address;
	natural_width guest_cr0;
	natural_width guest_cr3;
	natural_width guest_cr4;
	natural_width guest_es_base;
	natural_width guest_cs_base;
	natural_width guest_ss_base;
	natural_width guest_ds_base;
	natural_width guest_fs_base;
	natural_width guest_gs_base;
	natural_width guest_ldtr_base;
	natural_width guest_tr_base;
	natural_width guest_gdtr_base;
	natural_width guest_idtr_base;
	natural_width guest_dr7;
	natural_width guest_rsp;
	natural_width guest_rip;
	natural_width guest_rflags;
	natural_width guest_pending_dbg_exceptions;
	natural_width guest_sysenter_esp;
	natural_width guest_sysenter_eip;
	natural_width host_cr0;
	natural_width host_cr3;
	natural_width host_cr4;
	natural_width host_fs_base;
	natural_width host_gs_base;
	natural_width host_tr_base;
	natural_width host_gdtr_base;
	natural_width host_idtr_base;
	natural_width host_ia32_sysenter_esp;
	natural_width host_ia32_sysenter_eip;
	natural_width host_rsp;
	natural_width host_rip;
	natural_width paddingl[8]; /* room for future expansion */
	u32 pin_based_vm_exec_control;
	u32 cpu_based_vm_exec_control;
	u32 exception_bitmap;
	u32 page_fault_error_code_mask;
	u32 page_fault_error_code_match;
	u32 cr3_target_count;
	u32 vm_exit_controls;
	u32 vm_exit_msr_store_count;
	u32 vm_exit_msr_load_count;
	u32 vm_entry_controls;
	u32 vm_entry_msr_load_count;
	u32 vm_entry_intr_info_field;
	u32 vm_entry_exception_error_code;
	u32 vm_entry_instruction_len;
	u32 tpr_threshold;
	u32 secondary_vm_exec_control;
	u32 vm_instruction_error;
	u32 vm_exit_reason;
	u32 vm_exit_intr_info;
	u32 vm_exit_intr_error_code;
	u32 idt_vectoring_info_field;
	u32 idt_vectoring_error_code;
	u32 vm_exit_instruction_len;
	u32 vmx_instruction_info;
	u32 guest_es_limit;
	u32 guest_cs_limit;
	u32 guest_ss_limit;
	u32 guest_ds_limit;
	u32 guest_fs_limit;
	u32 guest_gs_limit;
	u32 guest_ldtr_limit;
	u32 guest_tr_limit;
	u32 guest_gdtr_limit;
	u32 guest_idtr_limit;
	u32 guest_es_ar_bytes;
	u32 guest_cs_ar_bytes;
	u32 guest_ss_ar_bytes;
	u32 guest_ds_ar_bytes;
	u32 guest_fs_ar_bytes;
	u32 guest_gs_ar_bytes;
	u32 guest_ldtr_ar_bytes;
	u32 guest_tr_ar_bytes;
	u32 guest_interruptibility_info;
	u32 guest_activity_state;
	u32 guest_sysenter_cs;
	u32 host_ia32_sysenter_cs;
	u32 vmx_preemption_timer_value;
	u32 padding32[7]; /* room for future expansion */
	u16 virtual_processor_id;
	u16 posted_intr_nv;
	u16 guest_es_selector;
	u16 guest_cs_selector;
	u16 guest_ss_selector;
	u16 guest_ds_selector;
	u16 guest_fs_selector;
	u16 guest_gs_selector;
	u16 guest_ldtr_selector;
	u16 guest_tr_selector;
	u16 guest_intr_status;
	u16 guest_pml_index;
	u16 host_es_selector;
	u16 host_cs_selector;
	u16 host_ss_selector;
	u16 host_ds_selector;
	u16 host_fs_selector;
	u16 host_gs_selector;
	u16 host_tr_selector;
};

struct pi_desc {
	u32 pir[8];     /* Posted interrupt requested */
	union {
		struct {
				/* bit 256 - Outstanding Notification */
			u16	on	: 1,
				/* bit 257 - Suppress Notification */
				sn	: 1,
				/* bit 271:258 - Reserved */
				rsvd_1	: 14;
				/* bit 279:272 - Notification Vector */
			u8	nv;
				/* bit 287:280 - Reserved */
			u8	rsvd_2;
				/* bit 319:288 - Notification Destination */
			u32	ndst;
		};
		u64 control;
	};
	u32 rsvd[6];
} __aligned(64);

struct vcpu_vmx {
	struct kvm_vcpu       vcpu;
	unsigned long         host_rsp;
	u8                    fail;
	u8		      msr_bitmap_mode;
	u32                   exit_intr_info;
	u32                   idt_vectoring_info;
	ulong                 rflags;
	struct shared_msr_entry *guest_msrs;
	int                   nmsrs;
	int                   save_nmsrs;
	unsigned long	      host_idt_base;
#ifdef CONFIG_X86_64
	u64 		      msr_host_kernel_gs_base;
	u64 		      msr_guest_kernel_gs_base;
#endif

	u64 		      arch_capabilities;
	u64 		      spec_ctrl;

	u32 vm_entry_controls_shadow;
	u32 vm_exit_controls_shadow;
	u32 secondary_exec_control;

	/*
	 * loaded_vmcs points to the VMCS currently used in this vcpu. For a
	 * non-nested (L1) guest, it always points to vmcs01. For a nested
	 * guest (L2), it points to a different VMCS.
	 */
	struct loaded_vmcs    vmcs01;
	struct loaded_vmcs   *loaded_vmcs;
	bool                  __launched; /* temporary, used in vmx_vcpu_run */
	struct msr_autoload {
		unsigned nr;
		struct vmx_msr_entry guest[KLP_NR_AUTOLOAD_MSRS];
		struct vmx_msr_entry host[KLP_NR_AUTOLOAD_MSRS];
	} msr_autoload;
	struct {
		int           loaded;
		u16           fs_sel, gs_sel, ldt_sel;
#ifdef CONFIG_X86_64
		u16           ds_sel, es_sel;
#endif
		int           gs_ldt_reload_needed;
		int           fs_reload_needed;
		u64           msr_host_bndcfgs;
	} host_state;
	struct {
		int vm86_active;
		ulong save_rflags;
		struct kvm_segment segs[8];
	} rmode;
	struct {
		u32 bitmask; /* 4 bits per segment (1 bit per field) */
		struct kvm_save_segment {
			u16 selector;
			unsigned long base;
			u32 limit;
			u32 ar;
		} seg[8];
	} segment_cache;
	int vpid;
	bool emulation_required;

	u32 exit_reason;

	/* Posted interrupt descriptor */
	struct pi_desc pi_desc;

	/* Support for a guest hypervisor (nested VMX) */
	struct nested_vmx nested;

	/* Dynamic PLE window. */
	int ple_window;
	bool ple_window_dirty;

	/* Support for PML */
#define PML_ENTITY_NUM		512
	struct page *pml_pg;

	/* apic deadline value in host tsc */
	u64 hv_deadline_tsc;

	u64 current_tsc_ratio;

	u32 host_pkru;

	/*
	 * Only bits masked by msr_ia32_feature_control_valid_bits can be set in
	 * msr_ia32_feature_control. FEATURE_CONTROL_LOCKED is always included
	 * in msr_ia32_feature_control_valid_bits.
	 */
	u64 msr_ia32_feature_control;
	u64 msr_ia32_feature_control_valid_bits;
};

/* inlined */
static inline struct vcpu_vmx *klp_to_vmx(struct kvm_vcpu *vcpu)
{
	return container_of(vcpu, struct vcpu_vmx, vcpu);
}

/* inlined */
static inline struct vmcs12 *klp_get_vmcs12(struct kvm_vcpu *vcpu)
{
	return klp_to_vmx(vcpu)->nested.cached_vmcs12;
}

/* inlined */
static inline unsigned klp_nested_cpu_vmx_misc_cr3_count(struct kvm_vcpu *vcpu)
{
	return vmx_misc_cr3_count(klp_to_vmx(vcpu)->nested.nested_vmx_misc_low);
}

/* inlined */
static inline bool klp_nested_cpu_has(struct vmcs12 *vmcs12, u32 bit)
{
	return vmcs12->cpu_based_vm_exec_control & bit;
}

/* inlined */
static inline bool klp_nested_cpu_has2(struct vmcs12 *vmcs12, u32 bit)
{
	return (vmcs12->cpu_based_vm_exec_control &
			CPU_BASED_ACTIVATE_SECONDARY_CONTROLS) &&
		(vmcs12->secondary_vm_exec_control & bit);
}

/* inlined */
static inline int klp_nested_cpu_has_ept(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_EPT);
}

/* inlined */
static inline bool klp_nested_cpu_has_virt_x2apic_mode(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUALIZE_X2APIC_MODE);
}

/* inlined */
static inline bool klp_nested_cpu_has_apic_reg_virt(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_APIC_REGISTER_VIRT);
}

/* inlined */
static inline bool klp_nested_cpu_has_vid(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUAL_INTR_DELIVERY);
}

/* inlined */
static inline bool klp_nested_cpu_has_posted_intr(struct vmcs12 *vmcs12)
{
	return vmcs12->pin_based_vm_exec_control & PIN_BASED_POSTED_INTR;
}

/* inlined */
static inline bool klp_nested_cpu_has_vmfunc(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_VMFUNC);
}

/* inlined */
static inline bool klp_nested_cpu_has_eptp_switching(struct vmcs12 *vmcs12)
{
	return klp_nested_cpu_has_vmfunc(vmcs12) &&
		(vmcs12->vm_function_control &
		 VMX_VMFUNC_EPTP_SWITCHING);
}

/* inlined */
static __always_inline void klp_vmcs_check16(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2000,
			 "16-bit accessor invalid for 64-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2001,
			 "16-bit accessor invalid for 64-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x4000,
			 "16-bit accessor invalid for 32-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x6000,
			 "16-bit accessor invalid for natural width field");
}

/* inlined */
static __always_inline void klp_vmcs_check32(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0,
			 "32-bit accessor invalid for 16-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x6000,
			 "32-bit accessor invalid for natural width field");
}

/* inlined */
static __always_inline void klp_vmcs_checkl(unsigned long field)
{
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0,
			 "Natural width accessor invalid for 16-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2000,
			 "Natural width accessor invalid for 64-bit field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6001) == 0x2001,
			 "Natural width accessor invalid for 64-bit high field");
        BUILD_BUG_ON_MSG(__builtin_constant_p(field) && ((field) & 0x6000) == 0x4000,
			 "Natural width accessor invalid for 32-bit field");
}

/* inlined */
static __always_inline unsigned long klp__vmcs_readl(unsigned long field)
{
	unsigned long value;

	asm volatile (klp__ex_clear(ASM_VMX_VMREAD_RDX_RAX, "%0")
		      : [thunk_target] "=a"(value) : "d"(field) : "cc");
	return value;
}

/* inlined */
static __always_inline u32 klp_vmcs_read32(unsigned long field)
{
	klp_vmcs_check32(field);
	return klp__vmcs_readl(field);
}

/* inlined */
static __always_inline void klp__vmcs_writel(unsigned long field,
					     unsigned long value)
{
	u8 error;

	asm volatile (klp__ex(ASM_VMX_VMWRITE_RAX_RDX) "; setna %0"
		       : "=q"(error) : [thunk_target] "a"(value), "d"(field) : "cc");
	if (unlikely(error))
		klp_vmwrite_error(field, value);
}

/* inlined */
static __always_inline void klp_vmcs_write16(unsigned long field, u16 value)
{
	klp_vmcs_check16(field);
	klp__vmcs_writel(field, value);
}

/* inlined */
static __always_inline void klp_vmcs_writel(unsigned long field,
					    unsigned long value)
{
	klp_vmcs_checkl(field);
	klp__vmcs_writel(field, value);
}

/* inlined */
static inline bool klp_fixed_bits_valid(u64 val, u64 fixed0, u64 fixed1)
{
	return ((val & fixed1) | fixed0) == val;
}

/* inlined */
static inline bool klp_vmx_control_verify(u32 control, u32 low, u32 high)
{
	return klp_fixed_bits_valid(control, low, high);
}

/* inlined */
static bool klp_nested_guest_cr0_valid(struct kvm_vcpu *vcpu, unsigned long val)
{
	u64 fixed0 = klp_to_vmx(vcpu)->nested.nested_vmx_cr0_fixed0;
	u64 fixed1 = klp_to_vmx(vcpu)->nested.nested_vmx_cr0_fixed1;
	struct vmcs12 *vmcs12 = klp_get_vmcs12(vcpu);

	if (klp_to_vmx(vcpu)->nested.nested_vmx_secondary_ctls_high &
		SECONDARY_EXEC_UNRESTRICTED_GUEST &&
	    klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_UNRESTRICTED_GUEST))
		fixed0 &= ~(X86_CR0_PE | X86_CR0_PG);

	return klp_fixed_bits_valid(val, fixed0, fixed1);
}

/* inlined */
static bool klp_nested_host_cr0_valid(struct kvm_vcpu *vcpu, unsigned long val)
{
	u64 fixed0 = klp_to_vmx(vcpu)->nested.nested_vmx_cr0_fixed0;
	u64 fixed1 = klp_to_vmx(vcpu)->nested.nested_vmx_cr0_fixed1;

	return klp_fixed_bits_valid(val, fixed0, fixed1);
}

/* inlined */
static bool klp_nested_cr4_valid(struct kvm_vcpu *vcpu, unsigned long val)
{
	u64 fixed0 = klp_to_vmx(vcpu)->nested.nested_vmx_cr4_fixed0;
	u64 fixed1 = klp_to_vmx(vcpu)->nested.nested_vmx_cr4_fixed1;

	return klp_fixed_bits_valid(val, fixed0, fixed1);
}

#define klp_nested_guest_cr4_valid	klp_nested_cr4_valid
#define klp_nested_host_cr4_valid	klp_nested_cr4_valid

/* inlined */
static bool klp_page_address_valid(struct kvm_vcpu *vcpu, gpa_t gpa)
{
	return PAGE_ALIGNED(gpa) && !(gpa >> klp_cpuid_maxphyaddr(vcpu));
}

/* inlined */
static bool klp_nested_exit_on_intr(struct kvm_vcpu *vcpu)
{
	return klp_get_vmcs12(vcpu)->pin_based_vm_exec_control &
		PIN_BASED_EXT_INTR_MASK;
}

/* inlined */
static bool klp_nested_exit_intr_ack_set(struct kvm_vcpu *vcpu)
{
	return klp_get_vmcs12(vcpu)->vm_exit_controls &
		VM_EXIT_ACK_INTR_ON_EXIT;
}

/* inlined */
static int klp_nested_vmx_check_vmcs12(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	if (vmx->nested.current_vmptr == -1ull) {
		klp_nested_vmx_failInvalid(vcpu);
		return 0;
	}
	return 1;
}

/* inlined */
static int klp_nested_vmx_check_io_bitmap_controls(struct kvm_vcpu *vcpu,
						   struct vmcs12 *vmcs12)
{
	if (!klp_nested_cpu_has(vmcs12, CPU_BASED_USE_IO_BITMAPS))
		return 0;

	if (!klp_page_address_valid(vcpu, vmcs12->io_bitmap_a) ||
	    !klp_page_address_valid(vcpu, vmcs12->io_bitmap_b))
		return -EINVAL;

	return 0;
}

/* inlined */
static int klp_nested_vmx_check_msr_bitmap_controls(struct kvm_vcpu *vcpu,
						    struct vmcs12 *vmcs12)
{
	if (!klp_nested_cpu_has(vmcs12, CPU_BASED_USE_MSR_BITMAPS))
		return 0;

	if (!klp_page_address_valid(vcpu, vmcs12->msr_bitmap))
		return -EINVAL;

	return 0;
}

/* inlined */
static int klp_nested_vmx_check_tpr_shadow_controls(struct kvm_vcpu *vcpu,
						    struct vmcs12 *vmcs12)
{
	if (!klp_nested_cpu_has(vmcs12, CPU_BASED_TPR_SHADOW))
		return 0;

	if (!klp_page_address_valid(vcpu, vmcs12->virtual_apic_page_addr))
		return -EINVAL;

	return 0;
}

/* inlined */
static int klp_nested_vmx_check_apicv_controls(struct kvm_vcpu *vcpu,
					       struct vmcs12 *vmcs12)
{
	if (!klp_nested_cpu_has_virt_x2apic_mode(vmcs12) &&
	    !klp_nested_cpu_has_apic_reg_virt(vmcs12) &&
	    !klp_nested_cpu_has_vid(vmcs12) &&
	    !klp_nested_cpu_has_posted_intr(vmcs12))
		return 0;

	/*
	 * If virtualize x2apic mode is enabled,
	 * virtualize apic access must be disabled.
	 */
	if (klp_nested_cpu_has_virt_x2apic_mode(vmcs12) &&
	    klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES))
		return -EINVAL;

	/*
	 * If virtual interrupt delivery is enabled,
	 * we must exit on external interrupts.
	 */
	if (klp_nested_cpu_has_vid(vmcs12) &&
	   !klp_nested_exit_on_intr(vcpu))
		return -EINVAL;

	/*
	 * bits 15:8 should be zero in posted_intr_nv,
	 * the descriptor address has been already checked
	 * in nested_get_vmcs12_pages.
	 */
	if (klp_nested_cpu_has_posted_intr(vmcs12) &&
	   (!klp_nested_cpu_has_vid(vmcs12) ||
	    !klp_nested_exit_intr_ack_set(vcpu) ||
	    vmcs12->posted_intr_nv & 0xff00))
		return -EINVAL;

	/* tpr shadow is needed by all apicv features. */
	if (!klp_nested_cpu_has(vmcs12, CPU_BASED_TPR_SHADOW))
		return -EINVAL;

	return 0;
}

/* inlined */
static int klp_nested_vmx_check_msr_switch_controls(struct kvm_vcpu *vcpu,
						    struct vmcs12 *vmcs12)
{
	if (vmcs12->vm_exit_msr_load_count == 0 &&
	    vmcs12->vm_exit_msr_store_count == 0 &&
	    vmcs12->vm_entry_msr_load_count == 0)
		return 0; /* Fast path */
	if (klp_nested_vmx_check_msr_switch(vcpu, VM_EXIT_MSR_LOAD_COUNT,
					    VM_EXIT_MSR_LOAD_ADDR) ||
	    klp_nested_vmx_check_msr_switch(vcpu, VM_EXIT_MSR_STORE_COUNT,
					    VM_EXIT_MSR_STORE_ADDR) ||
	    klp_nested_vmx_check_msr_switch(vcpu, VM_ENTRY_MSR_LOAD_COUNT,
					    VM_ENTRY_MSR_LOAD_ADDR))
		return -EINVAL;
	return 0;
}

/* inlined */
static int klp_nested_vmx_check_pml_controls(struct kvm_vcpu *vcpu,
					     struct vmcs12 *vmcs12)
{
	u64 address = vmcs12->pml_address;
	int maxphyaddr = klp_cpuid_maxphyaddr(vcpu);

	if (klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_PML)) {
		if (!klp_nested_cpu_has_ept(vmcs12) ||
		    !IS_ALIGNED(address, 4096)  ||
		    address >> maxphyaddr)
			return -EINVAL;
	}

	return 0;
}

/* inlined */
static bool klp_nested_cr3_valid(struct kvm_vcpu *vcpu, unsigned long val)
{
	unsigned long invalid_mask;

	invalid_mask = (~0ULL) << klp_cpuid_maxphyaddr(vcpu);
	return (val & invalid_mask) == 0;
}

/* inlined */
static int klp_check_vmentry_prereqs(struct kvm_vcpu *vcpu, struct vmcs12 *vmcs12)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);

	if (vmcs12->guest_activity_state != GUEST_ACTIVITY_ACTIVE &&
	    vmcs12->guest_activity_state != GUEST_ACTIVITY_HLT)
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_io_bitmap_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_msr_bitmap_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_tpr_shadow_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_apicv_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_msr_switch_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_vmx_check_pml_controls(vcpu, vmcs12))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (!klp_vmx_control_verify(vmcs12->cpu_based_vm_exec_control,
				    vmx->nested.nested_vmx_procbased_ctls_low,
				    vmx->nested.nested_vmx_procbased_ctls_high) ||
	    (klp_nested_cpu_has(vmcs12, CPU_BASED_ACTIVATE_SECONDARY_CONTROLS) &&
	     !klp_vmx_control_verify(vmcs12->secondary_vm_exec_control,
				     vmx->nested.nested_vmx_secondary_ctls_low,
				     vmx->nested.nested_vmx_secondary_ctls_high)) ||
	    !klp_vmx_control_verify(vmcs12->pin_based_vm_exec_control,
				    vmx->nested.nested_vmx_pinbased_ctls_low,
				    vmx->nested.nested_vmx_pinbased_ctls_high) ||
	    !klp_vmx_control_verify(vmcs12->vm_exit_controls,
				    vmx->nested.nested_vmx_exit_ctls_low,
				    vmx->nested.nested_vmx_exit_ctls_high) ||
	    !klp_vmx_control_verify(vmcs12->vm_entry_controls,
				    vmx->nested.nested_vmx_entry_ctls_low,
				    vmx->nested.nested_vmx_entry_ctls_high))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (klp_nested_cpu_has_vmfunc(vmcs12)) {
		if (vmcs12->vm_function_control &
		    ~vmx->nested.nested_vmx_vmfunc_controls)
			return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

		if (klp_nested_cpu_has_eptp_switching(vmcs12)) {
			if (!klp_nested_cpu_has_ept(vmcs12) ||
			    !klp_page_address_valid(vcpu, vmcs12->eptp_list_address))
				return VMXERR_ENTRY_INVALID_CONTROL_FIELD;
		}
	}

	if (vmcs12->cr3_target_count > klp_nested_cpu_vmx_misc_cr3_count(vcpu))
		return VMXERR_ENTRY_INVALID_CONTROL_FIELD;

	if (!klp_nested_host_cr0_valid(vcpu, vmcs12->host_cr0) ||
	    !klp_nested_host_cr4_valid(vcpu, vmcs12->host_cr4) ||
	    !klp_nested_cr3_valid(vcpu, vmcs12->host_cr3))
		return VMXERR_ENTRY_INVALID_HOST_STATE_FIELD;

	return 0;
}

/* inlined */
static int klp_check_vmentry_postreqs(struct kvm_vcpu *vcpu, struct vmcs12 *vmcs12,
				      u32 *exit_qual)
{
	bool ia32e;

	*exit_qual = ENTRY_FAIL_DEFAULT;

	if (!klp_nested_guest_cr0_valid(vcpu, vmcs12->guest_cr0) ||
	    !klp_nested_guest_cr4_valid(vcpu, vmcs12->guest_cr4))
		return 1;

	if (!klp_nested_cpu_has2(vmcs12, SECONDARY_EXEC_SHADOW_VMCS) &&
	    vmcs12->vmcs_link_pointer != -1ull) {
		*exit_qual = ENTRY_FAIL_VMCS_LINK_PTR;
		return 1;
	}

	/*
	 * If the load IA32_EFER VM-entry control is 1, the following checks
	 * are performed on the field for the IA32_EFER MSR:
	 * - Bits reserved in the IA32_EFER MSR must be 0.
	 * - Bit 10 (corresponding to IA32_EFER.LMA) must equal the value of
	 *   the IA-32e mode guest VM-exit control. It must also be identical
	 *   to bit 8 (LME) if bit 31 in the CR0 field (corresponding to
	 *   CR0.PG) is 1.
	 */
	if (klp_to_vmx(vcpu)->nested.nested_run_pending &&
	    (vmcs12->vm_entry_controls & VM_ENTRY_LOAD_IA32_EFER)) {
		ia32e = (vmcs12->vm_entry_controls & VM_ENTRY_IA32E_MODE) != 0;
		if (!klp_kvm_valid_efer(vcpu, vmcs12->guest_ia32_efer) ||
		    ia32e != !!(vmcs12->guest_ia32_efer & EFER_LMA) ||
		    ((vmcs12->guest_cr0 & X86_CR0_PG) &&
		     ia32e != !!(vmcs12->guest_ia32_efer & EFER_LME)))
			return 1;
	}

	/*
	 * If the load IA32_EFER VM-exit control is 1, bits reserved in the
	 * IA32_EFER MSR must be 0 in the field for that register. In addition,
	 * the values of the LMA and LME bits in the field must each be that of
	 * the host address-space size VM-exit control.
	 */
	if (vmcs12->vm_exit_controls & VM_EXIT_LOAD_IA32_EFER) {
		ia32e = (vmcs12->vm_exit_controls &
			 VM_EXIT_HOST_ADDR_SPACE_SIZE) != 0;
		if (!klp_kvm_valid_efer(vcpu, vmcs12->host_ia32_efer) ||
		    ia32e != !!(vmcs12->host_ia32_efer & EFER_LMA) ||
		    ia32e != !!(vmcs12->host_ia32_efer & EFER_LME))
			return 1;
	}

	if ((vmcs12->vm_entry_controls & VM_ENTRY_LOAD_BNDCFGS) &&
		(klp_is_noncanonical_address(vmcs12->guest_bndcfgs & PAGE_MASK, vcpu) ||
		(vmcs12->guest_bndcfgs & MSR_IA32_BNDCFGS_RSVD)))
			return 1;

	return 0;
}

/* inlined */
static void klp_nested_vmx_entry_failure(struct kvm_vcpu *vcpu,
			struct vmcs12 *vmcs12,
			u32 reason, unsigned long qualification)
{
	klp_load_vmcs12_host_state(vcpu, vmcs12);
	vmcs12->vm_exit_reason = reason | VMX_EXIT_REASONS_FAILED_VMENTRY;
	vmcs12->exit_qualification = qualification;
	klp_nested_vmx_succeed(vcpu);
	if (*klp_enable_shadow_vmcs)
		klp_to_vmx(vcpu)->nested.sync_shadow_vmcs = true;
}



/* patched */
void klp_vmx_save_host_state(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	int i;

	/*
	 * Fix CVE-2018-3646
	 *  +3 lines
	 *
	 * Note: we can't rely on the klp_enable_ept pointer being
	 * valid when executing in the context of kvm.ko. Hence we
	 * must either check it here or live with L1D flushes on
	 * !enable_ept hosts.
	 */
	if (!*klp_enable_ept)
		klp_get_and_clear_vcpu_unconfined(vcpu, GFP_ATOMIC);

	if (vmx->host_state.loaded)
		return;

	vmx->host_state.loaded = 1;
	/*
	 * Set host fs and gs selectors.  Unfortunately, 22.2.3 does not
	 * allow segment selectors with cpl > 0 or ti == 1.
	 */
	vmx->host_state.ldt_sel = kvm_read_ldt();
	vmx->host_state.gs_ldt_reload_needed = vmx->host_state.ldt_sel;
	savesegment(fs, vmx->host_state.fs_sel);
	if (!(vmx->host_state.fs_sel & 7)) {
		klp_vmcs_write16(HOST_FS_SELECTOR, vmx->host_state.fs_sel);
		vmx->host_state.fs_reload_needed = 0;
	} else {
		klp_vmcs_write16(HOST_FS_SELECTOR, 0);
		vmx->host_state.fs_reload_needed = 1;
	}
	savesegment(gs, vmx->host_state.gs_sel);
	if (!(vmx->host_state.gs_sel & 7))
		klp_vmcs_write16(HOST_GS_SELECTOR, vmx->host_state.gs_sel);
	else {
		klp_vmcs_write16(HOST_GS_SELECTOR, 0);
		vmx->host_state.gs_ldt_reload_needed = 1;
	}

#ifdef CONFIG_X86_64
	savesegment(ds, vmx->host_state.ds_sel);
	savesegment(es, vmx->host_state.es_sel);
#endif

#ifdef CONFIG_X86_64
	klp_vmcs_writel(HOST_FS_BASE, read_msr(MSR_FS_BASE));
	klp_vmcs_writel(HOST_GS_BASE, read_msr(MSR_GS_BASE));
#else
	klp_vmcs_writel(HOST_FS_BASE, segment_base(vmx->host_state.fs_sel));
	klp_vmcs_writel(HOST_GS_BASE, segment_base(vmx->host_state.gs_sel));
#endif

#ifdef CONFIG_X86_64
	rdmsrl(MSR_KERNEL_GS_BASE, vmx->msr_host_kernel_gs_base);
	if (klp_is_long_mode(&vmx->vcpu))
		wrmsrl(MSR_KERNEL_GS_BASE, vmx->msr_guest_kernel_gs_base);
#endif
	if (boot_cpu_has(X86_FEATURE_MPX))
		rdmsrl(MSR_IA32_BNDCFGS, vmx->host_state.msr_host_bndcfgs);
	for (i = 0; i < vmx->save_nmsrs; ++i)
		klp_kvm_set_shared_msr(vmx->guest_msrs[i].index,
				       vmx->guest_msrs[i].data,
				       vmx->guest_msrs[i].mask);
}

/* patched */
void klp_vmx_handle_external_intr(struct kvm_vcpu *vcpu)
{
	u32 exit_intr_info = klp_vmcs_read32(VM_EXIT_INTR_INFO);

	if ((exit_intr_info & (INTR_INFO_VALID_MASK | INTR_INFO_INTR_TYPE_MASK))
			== (INTR_INFO_VALID_MASK | INTR_TYPE_EXT_INTR)) {
		unsigned int vector;
		unsigned long entry;
		gate_desc *desc;
		struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
#ifdef CONFIG_X86_64
		unsigned long tmp;
#endif

		vector =  exit_intr_info & INTR_INFO_VECTOR_MASK;
		desc = (gate_desc *)vmx->host_idt_base + vector;
		entry = gate_offset(*desc);
		asm volatile(
#ifdef CONFIG_X86_64
			"mov %%" _ASM_SP ", %[sp]\n\t"
			"and $0xfffffffffffffff0, %%" _ASM_SP "\n\t"
			"push $%c[ss]\n\t"
			"push %[sp]\n\t"
#endif
			"pushf\n\t"
			__ASM_SIZE(push) " $%c[cs]\n\t"
			CALL_NOSPEC
			:
#ifdef CONFIG_X86_64
			[sp]"=&r"(tmp),
#endif
			ASM_CALL_CONSTRAINT
			:
			THUNK_TARGET(entry),
			[ss]"i"(__KERNEL_DS),
			[cs]"i"(__KERNEL_CS)
			);

		/*
		 * Fix CVE-2018-3646
		 *  +1 line
		 */
		klp_set_vcpu_unconfined(vcpu, GFP_ATOMIC);

	}
}
STACK_FRAME_NON_STANDARD(klp_vmx_handle_external_intr);

/* patched */
static int klp_nested_vmx_check_permission(struct kvm_vcpu *vcpu)
{
	if (klp_vmx_get_cpl(vcpu)) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 0;
	}

	if (!klp_to_vmx(vcpu)->nested.vmxon) {
		klp_kvm_queue_exception(vcpu, UD_VECTOR);
		return 0;
	}
	return 1;
}

/* patched */
int klp_nested_vmx_run(struct kvm_vcpu *vcpu, bool launch)
{
	struct vmcs12 *vmcs12;
	struct vcpu_vmx *vmx = klp_to_vmx(vcpu);
	u32 interrupt_shadow = klp_vmx_get_interrupt_shadow(vcpu);
	u32 exit_qual;
	int ret;

	if (!klp_nested_vmx_check_permission(vcpu))
		return 1;

	if (!klp_nested_vmx_check_vmcs12(vcpu))
		goto out;

	vmcs12 = klp_get_vmcs12(vcpu);

	if (*klp_enable_shadow_vmcs)
		klp_copy_shadow_to_vmcs12(vmx);

	/*
	 * The nested entry process starts with enforcing various prerequisites
	 * on vmcs12 as required by the Intel SDM, and act appropriately when
	 * they fail: As the SDM explains, some conditions should cause the
	 * instruction to fail, while others will cause the instruction to seem
	 * to succeed, but return an EXIT_REASON_INVALID_STATE.
	 * To speed up the normal (success) code path, we should avoid checking
	 * for misconfigurations which will anyway be caught by the processor
	 * when using the merged vmcs02.
	 */
	if (interrupt_shadow & KVM_X86_SHADOW_INT_MOV_SS) {
		klp_nested_vmx_failValid(vcpu,
					 VMXERR_ENTRY_EVENTS_BLOCKED_BY_MOV_SS);
		goto out;
	}

	if (vmcs12->launch_state == launch) {
		klp_nested_vmx_failValid(vcpu,
			launch ? VMXERR_VMLAUNCH_NONCLEAR_VMCS
			       : VMXERR_VMRESUME_NONLAUNCHED_VMCS);
		goto out;
	}

	ret = klp_check_vmentry_prereqs(vcpu, vmcs12);
	if (ret) {
		klp_nested_vmx_failValid(vcpu, ret);
		goto out;
	}

	/*
	 * After this point, the trap flag no longer triggers a singlestep trap
	 * on the vm entry instructions; don't call kvm_skip_emulated_instruction.
	 * This is not 100% correct; for performance reasons, we delegate most
	 * of the checks on host state to the processor.  If those fail,
	 * the singlestep trap is missed.
	 */
	klp_skip_emulated_instruction(vcpu);

	ret = klp_check_vmentry_postreqs(vcpu, vmcs12, &exit_qual);
	if (ret) {
		klp_nested_vmx_entry_failure(vcpu, vmcs12,
					     EXIT_REASON_INVALID_STATE, exit_qual);
		return 1;
	}

	/*
	 * We're finally done with prerequisite checking, and can start with
	 * the nested entry.
	 */

	ret = klp_enter_vmx_non_root_mode(vcpu, true);
	if (ret)
		return ret;

       /*
	* Fix CVE-2018-3646
	*  +3 lines
	*/
	/* Hide L1D cache contents from the nested guest. */
	klp_set_vcpu_unconfined(vcpu, GFP_KERNEL);

	if (vmcs12->guest_activity_state == GUEST_ACTIVITY_HLT)
		return klp_kvm_vcpu_halt(vcpu);

	vmx->nested.nested_run_pending = 1;

	return 1;

out:
	return klp_kvm_skip_emulated_instruction(vcpu);
}



static int livepatch_bsc1099306_kvm_intel_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(klp_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(klp_funcs[i].name);
		if (!addr) {
			pr_err("livepatch: symbol %s not resolved\n",
				klp_funcs[i].name);
			return -ENOENT;
		}

		*(klp_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int
livepatch_bsc1099306_kvm_intel_module_notify(struct notifier_block *nb,
					     unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KLP_PATCHED_MODULE))
		return 0;

	ret = livepatch_bsc1099306_kvm_intel_kallsyms();
	WARN(ret, "livepatch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block livepatch_bsc1099306_kvm_intel_module_nb = {
	.notifier_call = livepatch_bsc1099306_kvm_intel_module_notify,
	.priority = INT_MIN+1,
};

int __livepatch_bsc1099306_kvm_intel_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KLP_PATCHED_MODULE)) {
		ret = livepatch_bsc1099306_kvm_intel_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&livepatch_bsc1099306_kvm_intel_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void __livepatch_bsc1099306_kvm_intel_cleanup(void)
{
	unregister_module_notifier(&livepatch_bsc1099306_kvm_intel_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM) */
