/*
 * livepatch_bsc1103203
 *
 * Fix for CVE-2017-18344, bsc#1103203
 *
 *  Upstream commit:
 *  cef31d9af9082 ("posix-timer: Properly check sigevent->sigev_notify")
 *
 *  SLE12 commit:
 *  not affected (no CONFIG_CHECKPOINT_RESTORE, and thus, no show_timer())
 *
 *  SLE12-SP1 commit
 *  not affected (no CONFIG_CHECKPOINT_RESTORE, and thus, no show_timer())
 *
 *  SLE12-SP2 commit:
 *  90635b5f64786d5a7ad9121900a2fd08b94dbfad (stable 4.4.116)
 *
 *  SLE12-SP3 commit:
 *  8c5d5ef65b07d2a8f4288f01c3f25053a92a16f9 (stable 4.4.116)
 *
 *  SLE15 commit:
 *  2e3fc38c585099b2fed8cc3c1e3b658a0c643c18
 *
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/syscalls.h>
#include <linux/nospec.h>
#include <linux/slab.h>
#include <linux/posix-timers.h>
#include <linux/hashtable.h>
#include "livepatch_bsc1103203.h"
#include "kallsyms_relocs.h"


static struct k_clock *klp_clock_posix_dynamic;
static struct k_clock *klp_clock_posix_cpu;
static struct k_clock (*klp_posix_clocks)[MAX_CLOCKS];
static struct kmem_cache **klp_posix_timers_cache;
static struct hlist_head (*klp_posix_timers_hashtable)[1 << 9];
static spinlock_t *klp_hash_lock;

static struct task_struct *(*klp_find_task_by_vpid)(pid_t vnr);
static struct sigqueue *(*klp_sigqueue_alloc)(void);
static void (*klp_release_posix_timer)(struct k_itimer *tmr, int it_id_set);

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "clock_posix_dynamic", (void *)&klp_clock_posix_dynamic },
	{ "clock_posix_cpu", (void *)&klp_clock_posix_cpu },
	{ "posix_clocks", (void *)&klp_posix_clocks },
	{ "posix_timers_cache", (void *)&klp_posix_timers_cache },
	{ "posix_timers_hashtable", (void *)&klp_posix_timers_hashtable },
	{ "hash_lock", (void *)&klp_hash_lock, .sympos = 2 },
	{ "find_task_by_vpid", (void *)&klp_find_task_by_vpid },
	{ "sigqueue_alloc", (void *)&klp_sigqueue_alloc },
	{ "release_posix_timer", (void *)&klp_release_posix_timer },
};



/* from kernel/time/posix-timers.c */
/* inlined */
static int klp_hash(struct signal_struct *sig, unsigned int nr)
{
	return hash_32(hash32_ptr(sig) ^ nr, HASH_BITS(*klp_posix_timers_hashtable));
}

/* inlined */
static struct k_itimer *klp__posix_timers_find(struct hlist_head *head,
					       struct signal_struct *sig,
					       timer_t id)
{
	struct k_itimer *timer;

	hlist_for_each_entry_rcu(timer, head, t_hash) {
		if ((timer->it_signal == sig) && (timer->it_id == id))
			return timer;
	}
	return NULL;
}

/* inlined */
static int klp_posix_timer_add(struct k_itimer *timer)
{
	struct signal_struct *sig = current->signal;
	int first_free_id = sig->posix_timer_id;
	struct hlist_head *head;
	int ret = -ENOENT;

	do {
		spin_lock(klp_hash_lock);
		head = &(*klp_posix_timers_hashtable)[klp_hash(sig, sig->posix_timer_id)];
		if (!klp__posix_timers_find(head, sig, sig->posix_timer_id)) {
			hlist_add_head_rcu(&timer->t_hash, head);
			ret = sig->posix_timer_id;
		}
		if (++sig->posix_timer_id < 0)
			sig->posix_timer_id = 0;
		if ((sig->posix_timer_id == first_free_id) && (ret == -ENOENT))
			/* Loop over all possible ids completed */
			ret = -EAGAIN;
		spin_unlock(klp_hash_lock);
	} while (ret == -ENOENT);
	return ret;
}

/* inlined */
static struct k_itimer * klp_alloc_posix_timer(void)
{
	struct k_itimer *tmr;
	tmr = kmem_cache_zalloc(*klp_posix_timers_cache, GFP_KERNEL);
	if (!tmr)
		return tmr;
	if (unlikely(!(tmr->sigq = klp_sigqueue_alloc()))) {
		kmem_cache_free(*klp_posix_timers_cache, tmr);
		return NULL;
	}
	memset(&tmr->sigq->info, 0, sizeof(siginfo_t));
	return tmr;
}

#define KLP_IT_ID_SET		1
#define KLP_IT_ID_NOT_SET	0

/* partially inlined */
static struct k_clock *klp_clockid_to_kclock(const clockid_t id)
{
	clockid_t idx = id;

	if (id < 0) {
		return (id & CLOCKFD_MASK) == CLOCKFD ?
			klp_clock_posix_dynamic : klp_clock_posix_cpu;
	}

	if (id >= MAX_CLOCKS)
		return NULL;

	idx = array_index_nospec(idx, MAX_CLOCKS);

	if (!(*klp_posix_clocks)[idx].clock_getres)
		return NULL;

	return &(*klp_posix_clocks)[idx];
}



/* patched, inlined */
static struct pid *klp_good_sigevent(sigevent_t * event)
{
	struct task_struct *rtn = current->group_leader;

	/*
	 * Fix CVE-2017-18344
	 *  -11 lines (all), +16 lines
	 */
	switch (event->sigev_notify) {
	case SIGEV_SIGNAL | SIGEV_THREAD_ID:
		rtn = klp_find_task_by_vpid(event->sigev_notify_thread_id);
		if (!rtn || !same_thread_group(rtn, current))
			return NULL;
		/* FALLTHRU */
	case SIGEV_SIGNAL:
	case SIGEV_THREAD:
		if (event->sigev_signo <= 0 || event->sigev_signo > SIGRTMAX)
			return NULL;
		/* FALLTHRU */
	case SIGEV_NONE:
		return task_pid(rtn);
	default:
		return NULL;
	}
}

/* patched, calls good_sigevent() */
__SYSCALL_DEFINEx(3, _klp_timer_create, const clockid_t, which_clock,
		  struct sigevent __user *, timer_event_spec,
		  timer_t __user *, created_timer_id)
{
	struct k_clock *kc = klp_clockid_to_kclock(which_clock);
	struct k_itimer *new_timer;
	int error, new_timer_id;
	sigevent_t event;
	int it_id_set = KLP_IT_ID_NOT_SET;

	if (!kc)
		return -EINVAL;
	if (!kc->timer_create)
		return -EOPNOTSUPP;

	new_timer = klp_alloc_posix_timer();
	if (unlikely(!new_timer))
		return -EAGAIN;

	spin_lock_init(&new_timer->it_lock);
	new_timer_id = klp_posix_timer_add(new_timer);
	if (new_timer_id < 0) {
		error = new_timer_id;
		goto out;
	}

	it_id_set = KLP_IT_ID_SET;
	new_timer->it_id = (timer_t) new_timer_id;
	new_timer->it_clock = which_clock;
	new_timer->it_overrun = -1;

	if (timer_event_spec) {
		if (copy_from_user(&event, timer_event_spec, sizeof (event))) {
			error = -EFAULT;
			goto out;
		}
		rcu_read_lock();
		new_timer->it_pid = get_pid(klp_good_sigevent(&event));
		rcu_read_unlock();
		if (!new_timer->it_pid) {
			error = -EINVAL;
			goto out;
		}
	} else {
		memset(&event.sigev_value, 0, sizeof(event.sigev_value));
		event.sigev_notify = SIGEV_SIGNAL;
		event.sigev_signo = SIGALRM;
		event.sigev_value.sival_int = new_timer->it_id;
		new_timer->it_pid = get_pid(task_tgid(current));
	}

	new_timer->it_sigev_notify     = event.sigev_notify;
	new_timer->sigq->info.si_signo = event.sigev_signo;
	new_timer->sigq->info.si_value = event.sigev_value;
	new_timer->sigq->info.si_tid   = new_timer->it_id;
	new_timer->sigq->info.si_code  = SI_TIMER;

	if (copy_to_user(created_timer_id,
			 &new_timer_id, sizeof (new_timer_id))) {
		error = -EFAULT;
		goto out;
	}

	error = kc->timer_create(new_timer);
	if (error)
		goto out;

	spin_lock_irq(&current->sighand->siglock);
	new_timer->it_signal = current->signal;
	list_add(&new_timer->list, &current->signal->posix_timers);
	spin_unlock_irq(&current->sighand->siglock);

	return 0;
	/*
	 * In the case of the timer belonging to another task, after
	 * the task is unlocked, the timer is owned by the other task
	 * and may cease to exist at any time.  Don't use or modify
	 * new_timer after the unlock call.
	 */
out:
	klp_release_posix_timer(new_timer, it_id_set);
	return error;
}
/* alias for usedsymbols */
asmlinkage long klp_sys_timer_create(long which_clock, long timer_event_spec,
				     long created_timer_id)
	__attribute__((alias("SyS_klp_timer_create")));



int livepatch_bsc1103203_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
